
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "CBP.h"

static void	Error(msg)
char		*msg;
{
	fputs(msg,stderr); fputc('\n',stderr); exit(1);
}

static double	*Read1DProjection(NpS,dS,file)
int		*NpS;
double		*dS;
FILE		*file;
{
	double	*pS;
	int	i;

	fscanf(file,"%d",NpS);
	if ((pS=(double *)malloc(*NpS*sizeof(*pS)))==NULL) return NULL;

	fscanf(file,"%lf",dS);
	for (i=0; i<*NpS; i++) fscanf(file,"%lf",&(pS[i]));

	return pS;
}

static double	Interpolate1DProjection(NpS,dS,pS,S)
int		NpS;
double		dS,*pS,S;
{
	int	i,NpS2=NpS-2;

	i=(int)(S/dS); if (i>NpS2) i=NpS2;

	return pS[i]+(pS[i+1]-pS[i])*(S/dS-(double)i);
}

int	main(argc,argv)
int	argc;
char	**argv;
{
	FILE	*file;
	int	NpS,N,M,grids,i,regriding=0;
	double	dS,*pS,delta,radius,angle,Cos,Sin,r0,r,S,step;
	CBP	*cbp;

	if (argc!=7 && argc!=8)
	Error("usage : bhcbp 1DProjectionFile delta N {M} radius grids angle");

	if (argv[1][0]=='-')
	    file=stdin;
	else
	    if ((file=fopen(argv[1],"r"))==NULL) Error("file not found.");

	if ((pS=Read1DProjection(&NpS,&dS,file))==NULL)
	    Error("memory allocation error for 1D projection.");

	if (file!=stdin) fclose(file);

	if ((delta=atof(argv[2]))<=0.0) Error("bad delta.");

	if ((N=atoi(argv[3]))<=0) Error("bad N.");

	if (argc==7)
	    M=(int)ceil(M_PI*(double)N/2.0);
	else {
	    if ((M=atoi(argv[4]))==0) Error("bad M.");

	    if (M<0) {
		M=(-M); regriding=1;
	    }
	}

	if ((radius=atof(argv[argc-3]))<=0.0) Error("bad radius.");

	if ((grids=atoi(argv[argc-2]))<=1) Error("bad number of grids.");

	angle=atof(argv[argc-1]); Cos=cos(angle*M_PI/180.0);
				  Sin=sin(angle*M_PI/180.0);

	r0=delta*(double)(1-N)/2.0;

	if ((cbp=SetUpCBP(N,M,delta,r0,0.0))==NULL)
	    Error("memory allocation error for CBP.");

	for (i=0; i<N; i++) {
	    r=r0+delta*(double)i;
	    if (r>(-radius) && r<radius) {
		S=2.0*sqrt(radius*radius-r*r);
		cbp->pq[0][i]=Interpolate1DProjection(NpS,dS,pS,S);
	    }
	}
#ifndef	CONCENTRIC_SYMMETRY
{
	int	m;

	for (m=1; m<M; m++) for (i=0; i<N; i++) cbp->pq[m][i]=cbp->pq[0][i];
}
#endif
	Convolution(cbp);

	if (regriding)
	    if (!ReGridingCP(cbp)) Error("memory allocation error for CP.");

	step=r0/(double)(1-grids); printf("%d\t%lf\n",grids,step);

	for (i=0; i<grids; i++) {
	    r=step*(double)i; printf("%lf\n",BackProjection(cbp,r*Cos,r*Sin));
	}
	return 0;
}
