
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "ctd.h"

typedef unsigned char	uChar;
typedef unsigned short	uShort;
typedef unsigned long	uLong;

static void	CopyN(header,pos,ptr,size)
uChar		*header,*ptr;
int		pos,size;
{
	for (header+=pos; size>0; size--) *(ptr++)=(*(header++));
}

static void	CopyR(header,pos,ptr,size)
uChar		*header,*ptr;
int		pos,size;
{
	for (header+=(pos+size); size>0; size--) *(ptr++)=(*(--header));
}

static size_t	fRead(ptr,size,cnt,file)
uChar		*ptr;
size_t		size,cnt;
FILE		*file;
{
	size_t	i,j,size1=size-1,size2=size/2,read=fread(ptr,size,cnt,file);
	uChar	*p1,*p2,p;

	for (i=0; i<read; i++) {
	    for (j=0; j<size2; j++) {
		p1=ptr+j; p2=ptr+size1-j; p=(*p1); *p1=(*p2); *p2=p;
	    }
	    ptr+=size;
	}
	return read;
}

static int	GetS(file,Read,data)
FILE		*file;
size_t		(*Read)();
Float		*data;
{
	uShort	tmp;

	if ((*Read)(&tmp,sizeof(tmp),1,file)!=1) return 0;

	*data=(Float)tmp; return 1;
}

static int	GetL(file,Read,data)
FILE		*file;
size_t		(*Read)();
Float		*data;
{
	uLong	tmp;

	if ((*Read)(&tmp,sizeof(tmp),1,file)!=1) return 0;

	*data=(Float)tmp; return 1;
}

static int	ScanData(file,Read,Get,Nu,Ma,p)
FILE		*file;
size_t		(*Read)();
int		(*Get)();
short		Nu,Ma;
Float		**p;
{
	int	a,u;

	for (a=0; a<Ma; a++)
	for (u=0; u<Nu; u++) if (!(*Get)(file,Read,&(p[a][u]))) return 0;

	return 1;
}

static int	ScanMean(file,Read,Get,Nu,Mid,ID)
FILE		*file;
size_t		(*Read)();
int		(*Get)();
short		Nu,Mid;
Float		*ID;
{
	int	a,u;
	Float	tmp;

	for (a=0; a<Mid; a++)
	for (u=0; u<Nu ; u++) {
	    if (!(*Get)(file,Read,&tmp)) return 0;

	    ID[u]+=tmp;
	}
	for (u=0; u<Nu; u++) ID[u]/=(Float)Mid;

	return 1;
}

#define HEADER	ctd->header
#define NU	(ctd->Nu)
#define MA	(ctd->Ma)
#define P	(ctd->p)

CTD	*ReadCTD(path)
char	*path;
{
	uShort	word=0x1234;
	uChar	*byte=(uChar *)&word;
	int	(*Get)(),a,u,NR=(*byte==(word&0xff));
	void	(*Copy)()=(NR)?CopyN:CopyR;
	size_t	(*Read)()=(NR)?fread:fRead;
	CTD	*ctd=(CTD *)malloc(sizeof(*ctd));
	FILE	*file;
	short	Mi,Md;
	char	*pmae="projection memory allocation error.";
	Float	*I,*D,logI;

	if (ctd==NULL) {
	    fprintf(stderr,"%s : header memory allocation error.\n",path);
	    return NULL;
	}
	if (*path=='-')
	    file=stdin;
	else
	    if ((file=fopen(path,"rb"))==NULL) {
		fprintf(stderr,"%s : file not open.\n",path);
		free(ctd); return NULL;
	    }

	if (fread(HEADER,sizeof(HEADER),1,file)!=1) {
	    fprintf(stderr,"%s : header read error.\n",path);
	    if (file!=stdin) fclose(file); free(ctd); return NULL;
	}
	switch(HEADER[65-1]) {
	    case 0 : Get=GetS; break;
	    case 1 : Get=GetL; break;
	    default:
		fprintf(stderr,"%s : unknown data type.\n",path);
		if (file!=stdin) fclose(file); free(ctd); return NULL;
	}
	(*Copy)(HEADER,529-1,&(ctd->Ld),sizeof(ctd->Ld));
	(*Copy)(HEADER,533-1,&(ctd->Lc),sizeof(ctd->Lc));
	(*Copy)(HEADER,537-1,&(ctd->U ),sizeof(ctd->U ));
	(*Copy)(HEADER,527-1,&      NU ,sizeof(     NU));
	(*Copy)(HEADER,557-1,&      MA ,sizeof(     MA));
	(*Copy)(HEADER,567-1,&(ctd->Mt),sizeof(ctd->Mt));
	(*Copy)(HEADER,569-1,&      Mi ,sizeof(     Mi));
	(*Copy)(HEADER,571-1,&      Md ,sizeof(     Md));
	if (ctd->Ld<=0.0 || ctd->Lc<=0.0 || ctd->U <=0.0 ||
	    NU<=0 || MA<=0 || ctd->Mt<=0 || Mi<=0 || Md<=0) {
	    fprintf(stderr,"%s : bad parameter value.\n",path);
	    if (file!=stdin) fclose(file); free(ctd); return NULL;
	}
	if ((P=(Float **)malloc((int)MA*sizeof(*P)))==NULL) {
	    fprintf(stderr,"%s : %s\n",path,pmae);
	    if (file!=stdin) fclose(file); free(ctd); return NULL;
	}
	for (a=0; a<MA; a++)
	    if ((P[a]=(Float *)malloc((int)NU*sizeof(**P)))==NULL) {
		fprintf(stderr,"%s : %s\n",path,pmae);
		while (--a>=0) free(P[a]); free(P);
		if (file!=stdin) fclose(file); free(ctd); return NULL;
	    }

	if ((I=(Float *)malloc((int)NU*sizeof(*I)))==NULL) {
	    fprintf(stderr,"%s : %s\n",path,pmae);
	    for (a=MA-1; a>=0; a--) free(P[a]); free(P);
	    if (file!=stdin) fclose(file); free(ctd); return NULL;
	}
	if ((D=(Float *)malloc((int)NU*sizeof(*D)))==NULL) {
	    fprintf(stderr,"%s : %s\n",path,pmae);
	    free(I); for (a=MA-1; a>=0; a--) free(P[a]); free(P);
	    if (file!=stdin) fclose(file); free(ctd); return NULL;
	}
	for (u=0; u<NU; u++) I[u]=D[u]=0.0;

#ifdef	OLD
	if (!ScanData(file,Read,Get,NU,MA,P) ||
	    !ScanMean(file,Read,Get,NU,Mi,I) ||
	    !ScanMean(file,Read,Get,NU,Md,D))
#else
	if ((HEADER[601-1])?!ScanMean(file,Read,Get,NU,Md,D) ||
			    !ScanMean(file,Read,Get,NU,Mi,I) ||
			    !ScanData(file,Read,Get,NU,MA,P)
			   :!ScanData(file,Read,Get,NU,MA,P) ||
			    !ScanMean(file,Read,Get,NU,Mi,I) ||
			    !ScanMean(file,Read,Get,NU,Md,D))
#endif
	{
	    fprintf(stderr,"%s : data read error.\n",path);
	    free(D); free(I); for (a=MA-1; a>=0; a--) free(P[a]); free(P);
	    if (file!=stdin) fclose(file); free(ctd); return NULL;
	}
	if (file!=stdin) fclose(file);

	for (u=0; u<NU; u++)
	    if (I[u]<=D[u])
		for (a=0; a<MA; a++) P[a][u]=0.0;
	    else {
		logI=log(I[u]-D[u]);
		for (a=0; a<MA; a++)
		    P[a][u]=(P[a][u]<=D[u])?0.0:logI-log(P[a][u]-D[u]);
	    }

	free(D); free(I); return ctd;
}
