
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define STR_LEN	256

static char	text[STR_LEN];
static int	line;

static void	Abort(msg)
char		*msg;
{
	fprintf(stderr,"%s at line %d : %s\n",msg,line,text); exit(1);
}

typedef	unsigned char	uChar;
typedef	unsigned short	uShort;
typedef	unsigned long	uLong;

#define TYPE_CHAR	0
#define TYPE_UCHAR	1
#define TYPE_SHORT	2
#define TYPE_USHORT	3
#define TYPE_LONG	4
#define TYPE_ULONG	5
#define TYPE_FLOAT	6
#define TYPE_DOUBLE	7
#define TYPE_STRING	8
#define TYPE_O		9
#define TYPE_X		10

static struct {
		char	*key;		/* keyword */
		int	size;		/* data size */
		char	*format;	/* print format */
	} rule[]={
		{"c", 1,"%d"  },	/* 0 */
		{"uc",1,"%u"  },	/* 1 */
		{"sh",2,"%d"  },	/* 2 */
		{"us",2,"%u"  },	/* 3 */
		{"l", 4,"%d"  },	/* 4 */
		{"ul",4,"%lu" },	/* 5 */
		{"f", 4,"%f"  },	/* 6 */
		{"d", 8,"%lf" },	/* 7 */
		{"st",1,"%s"  },	/* 8 */
		{"o", 1,"%03o"},	/* 9 */
		{"x", 1,"%02x"}};	/* 10 */

#define CHECK_KEY(TYPE)	\
	if (!strncasecmp(key,rule[TYPE].key,strlen(rule[TYPE].key))) {\
	    *type=TYPE; return;\
	}

#ifdef	WITH_COMMENT
static void	CheckScript(loc,type,num,com)
int		*loc,*type,*num;
char		*com;
#else
static void	CheckScript(loc,type,num)
int		*loc,*type,*num;
#endif
{
	char	*ptr,key[STR_LEN];

	if ((ptr=strchr(text,'#'))==NULL) {
#ifdef	WITH_COMMENT
	    strcpy(com,"\n");
#endif
	    ptr=text+strlen(text)-1;
	}
#ifdef	WITH_COMMENT
	else
	    strcpy(com,ptr+1);
#endif
	do {
	    *ptr='\0'; if (--ptr<text) return;
	} while (*ptr==' ' || *ptr=='\t');

	switch(sscanf(text,"%d %s %d",loc,key,num)) {
	    case 1  : Abort("too few script item");
	    case 2  : *num=1;
	    case 3  : break;
	    default : Abort("script items scan error");
	}
	CHECK_KEY(TYPE_CHAR  );
	CHECK_KEY(TYPE_UCHAR );
	CHECK_KEY(TYPE_SHORT );
	CHECK_KEY(TYPE_USHORT);
	CHECK_KEY(TYPE_LONG  );
	CHECK_KEY(TYPE_ULONG );
	CHECK_KEY(TYPE_FLOAT );
	CHECK_KEY(TYPE_DOUBLE);
	CHECK_KEY(TYPE_STRING);
	CHECK_KEY(TYPE_O     );
	CHECK_KEY(TYPE_X     ); Abort("unknown data type");
}

#ifdef	WITH_COMMENT
#define PRINT_DATA(TYPE)	\
	for (idx=0; idx<num; idx++) {\
	    if (idx>0) putchar(' '); printf(format,*((TYPE *)(data+size*idx)));\
	}

static void	PrintCTD(CTD,fRead,loc,type,num,com)
FILE		*CTD;
size_t		(*fRead)();
int		loc,type,num;
char		*com;
#else
#define PRINT_DATA(TYPE)	\
	for (idx=0; idx<num; idx++) {\
	    printf(format,*((TYPE *)(data+size*idx))); putchar('\n');\
	}

static void	PrintCTD(CTD,fRead,loc,type,num)
FILE		*CTD;
size_t		(*fRead)();
int		loc,type,num;
#endif
{
	int	idx,size=rule[type].size;
	uChar	*data=(uChar *)malloc(size*num+((type==TYPE_STRING)?1:0));
	char	*format=rule[type].format;

	if (data==NULL) Abort("data memory allocation error");

	if (fseek(CTD,loc-1,0)!=0) Abort("bad data location");

	if ((*fRead)(data,size,num,CTD)!=num) Abort("bad data number");

	switch(type) {
	    case TYPE_CHAR   : PRINT_DATA(char  ); break;
	    case TYPE_UCHAR  : PRINT_DATA(uChar ); break;
	    case TYPE_SHORT  : PRINT_DATA(short ); break;
	    case TYPE_USHORT : PRINT_DATA(uShort); break;
	    case TYPE_LONG   : PRINT_DATA(long  ); break;
	    case TYPE_ULONG  : PRINT_DATA(uLong ); break;
	    case TYPE_FLOAT  : PRINT_DATA(float ); break;
	    case TYPE_DOUBLE : PRINT_DATA(double); break;
	    case TYPE_STRING :
		data[num]='\0'; printf(format,data);
#ifndef	WITH_COMMENT
		putchar('\n');
#endif
		break;
	    case TYPE_O      :
	    case TYPE_X      :
#ifdef	WITH_COMMENT
		PRINT_DATA(uChar);
#else
		for (idx=0; idx<num; idx++) {
		    if (idx>0) putchar(' '); printf(format,data[idx]);
		}
		putchar('\n');
#endif
		break;
	}
#ifdef	WITH_COMMENT
	printf("\t#%s",com);
#endif
	free(data);
}

static size_t	fReadRev(ptr,size,items,file)
char		*ptr;
size_t		size,items;
FILE		*file;
{
	size_t	i,j,size1=size-1,size2=size/2,read=fread(ptr,size,items,file);
	char	*p1,*p2,c;

	for (i=0; i<read; i++) {
	    for (j=0; j<size2; j++) {
		p1=ptr+j; p2=ptr+size1-j; c=(*p1); *p1=(*p2); *p2=c;
	    }
	    ptr+=size;
	}
	return read;
}

int	main(argc,argv)
int	argc;
char	**argv;
{
	uShort	tmp=0x1234;
	uChar	*ptr=(uChar *)&tmp;
	size_t	(*fRead)()=(*ptr==(tmp&0xff))?fread:fReadRev;
	FILE	*script,*CTD;
	int	loc,type,num;
#ifdef	WITH_COMMENT
	char	com[STR_LEN];
#endif

	if (argc!=2 && argc!=3) {
	    fprintf(stderr,"usage : %s {script} CTD\n",*argv); exit(1);
	}

	if (argc==2)
	    script=stdin;
	else
	    if ((script=fopen(argv[1],"r"))==NULL) {
		fprintf(stderr,"%s : script file not open.\n",argv[1]); exit(1);
	    }

	if ((CTD=fopen(argv[argc-1],"r"))==NULL) {
	    fprintf(stderr,"%s : CTD file not open.\n",argv[argc-1]); exit(1);
	}

	for (line=1; fgets(text,STR_LEN,script)!=NULL; line++)
	    if (text[strlen(text)-1]!='\n') {
		fprintf(stderr,"too long script at line %d\n",line); exit(1);
	    }
	    else {
#ifdef	WITH_COMMENT
		CheckScript(&loc,&type,&num,com);
		if (*text!='\0') PrintCTD(CTD,fRead,loc,type,num,com);
#else
		CheckScript(&loc,&type,&num);
		if (*text!='\0') PrintCTD(CTD,fRead,loc,type,num);
#endif
	    }

	fclose(CTD); if (script!=stdin) fclose(script); return 0;
}
