
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "ctd.h"

static void	Error(msg)
char		*msg;
{
	fputs(msg,stderr); fputc('\n',stderr); exit(1);
}

#define ALLOC(num,ptr)	malloc((size_t)(num)*sizeof(*(ptr)))

int	main(argc,argv)
int	argc;
char	**argv;
{
	CTD	*ctd;
	double	Ld,Lc,mf,du,a0,da,us,uc,dr,dt,rc,**p1,
		ucLd,Ld2,sa,dth,u0,r0,u1,r1,su,a1,u2,t0,r2,t1,r3,p;
	int	Nr,Ng,Mg,**p0,t,r,u,a,ng,n,m;
	char	*pmae="projection memory allocatoin error.",
		*fns="file not stored.";
	FILE	*file;

	if (argc!=8 && argc!=9)
	Error("usage : ctd2prt CTD Nr Ng Mg dUs dUc binaryArray {ASCIIArray}");

	if ((ctd=ReadCTD(argv[1]))==NULL) exit(1);

	Ld=ctd->Ld;
	Lc=ctd->Lc;
	mf=Ld/Lc;
	du=ctd->U/(double)ctd->Nu;
	a0=M_PI/2.0+atan((ctd->U/2.0)/Ld);
	da=(-M_PI)/(double)ctd->Mt;

	fprintf(stderr,"%lf\t%lf\n%d\t%d\n",mf,du,ctd->Nu,ctd->Ma);

	if ((Nr=atoi(argv[2]))<=0 ||
	    (Ng=atoi(argv[3]))<=0 ||
	    (Mg=atoi(argv[4]))<=0) Error("bad grid number.");

	us=du*((double)ctd->Nu/2.0+atof(argv[5]));
	uc=du*atof(argv[6]);
	dr=ctd->U/(mf*(double)Nr);
	rc=dr*(double)Nr/2.0;
	dt=M_PI/(double)ctd->Mt;

	fprintf(stderr,"%lf\n",dr);

	if ((p0=(int    **)ALLOC(ctd->Mt,p0))==NULL ||
	    (p1=(double **)ALLOC(ctd->Mt,p1))==NULL) Error(pmae);

	for (t=0; t<ctd->Mt; t++) {
	    if ((p0[t]=(int    *)ALLOC(Nr,*p0))==NULL ||
		(p1[t]=(double *)ALLOC(Nr,*p1))==NULL) Error(pmae);

	    for (r=0; r<Nr; r++) {
		p0[t][r]=0; p1[t][r]=0.0;
	}   }

	ucLd=uc*Ld;
	Ld2=Ld*Ld;
	sa=da/(double)Mg;
	dth=dt/2.0;

	u0=(-us); r0=(u0*Lc-ucLd)/sqrt(u0*u0+Ld2);
	for (u=0; u<ctd->Nu; u++) {
	    u1=du*(double)(u+1)-us; r1=(u1*Lc-ucLd)/sqrt(u1*u1+Ld2);
	    ng=Ng*(int)(ceil((r1+rc)/dr)-floor((r0+rc)/dr));
	    su=du/(double)ng;
	    for (a=0; a<ctd->Ma; a++) {
		a1=a0+da*((double)a-0.5);
		p=ctd->p[a][u];
		for (n=0; n<ng; n++) {
		    u2=u0+su*((double)n+0.5); r2=(u2*Lc-ucLd)/sqrt(u2*u2+Ld2);
		    t0=a1-atan(u2/Ld);
		    for (m=0; m<Mg; m++) {
			t1=t0+sa*((double)m+0.5);
			if (t1<-dth) {
			    r3=(-r2); t1+=M_PI;
			}
			else
			    r3=r2;

			r=(int)floor((r3+rc)/dr);
			if (r>=0 && r<Nr) {
			    t=(int)floor((t1+dth)/dt);
			    ++p0[t][r]; p1[t][r]+=p;
	    }   }   }   }

	    u0=u1; r0=r1;
	}
	if (argv[7][0]=='-')
	    file=stdout;
	else
	    if ((file=fopen(argv[7],"wb"))==NULL) Error(fns);

	fprintf(file,"%d\t%d\n",Nr,ctd->Mt);
	for (t=ctd->Mt-1; t>=0; t--)
	for (r=     Nr-1; r>=0; r--) {
	    p=(p0[t][r]>0)?p1[t][r]/(dr*(double)p0[t][r]):0.0;
	    fwrite(&p,sizeof(p),1,file);
	}
	if (file!=stdout) fclose(file);

	if (argc==9) {
	    if (argv[8][0]=='-')
		file=stdout;
	    else
		if ((file=fopen(argv[8],"wb"))==NULL) Error(fns);

	    fprintf(file,"%d\t%d\n",Nr,ctd->Mt);
	    for (t=ctd->Mt-1; t>=0; t--)
	    for (r=     Nr-1; r>=0; r--) fprintf(file,"%d\n",p0[t][r]);

	    if (file!=stdout) fclose(file);
	}
	return 0;
}
