
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "cell.h"
#include "sif.h"

static void	Error(msg)
char		*msg;
{
	fputs(msg,stderr); fputc('\n',stderr); exit(1);
}

static size_t	fReadRev(ptr,size,items,file)
char		*ptr;
size_t		size,items;
FILE		*file;
{
	size_t	i,j,size1=size-1,size2=size/2,read=fread(ptr,size,items,file);
	char	*p1,*p2,c;

	for (i=0; i<read; i++) {
	    for (j=0; j<size2; j++) {
		p1=ptr+j; p2=ptr+size1-j; c=(*p1); *p1=(*p2); *p2=c;
	    }
	    ptr+=size;
	}
	return read;
}

static unsigned short	*sData=NULL;
static unsigned long	*lData=NULL;
static int		idx,lim;
static double		base,step;

static Cell	GetCellS()
{
	double	d=((double)sData[idx++]-base)/step;

	return (d<0.0)?0:(d>=(double)lim)?(Cell)lim:(Cell)floor(d+0.5);
}

static Cell	GetCellL()
{
	double	d=((double)lData[idx++]-base)/step;

	return (d<0.0)?0:(d>=(double)lim)?(Cell)lim:(Cell)floor(d+0.5);
}

#define DESC_LEN	256
#define	MA(num,var)	malloc((size_t)(num)*sizeof(*(var)))

int	main(argc,argv)
int	argc;
char	**argv;
{
	unsigned short	*s,tmp=0x1234;
	unsigned char	*ptr=(unsigned char *)&tmp;
	size_t		(*fRead)()=(*ptr==(tmp&0xff))?fread:fReadRev;
	FILE		*file;
	int		BPS,type,total,read,hight,x,y,loc=1;
	Cell		(*GetCell)();
	short		width,I,I0,dark;
	float		DSU;
	double		dr;
	unsigned long	*l,min,max;
	char		desc[DESC_LEN];

	if (argc!=4 && argc!=6)
	    Error("usage : ctd2tiff CTD {base step} BPS TIFF");

	if (argv[1][0]=='-')
	    file=stdin;
	else
	    if ((file=fopen(argv[1],"rb"))==NULL) {
		fputs(argv[1],stderr); Error(" : file not open.");
	    }

	if (argc==6) {
	    base=atof(argv[2]);
	    if ((step=atof(argv[3]))<=0.0) Error("bad image data step.");
	}
	if ((BPS=atoi(argv[argc-2]))<=0 || BPS>8*sizeof(Cell))
	    Error("bad image cell BPS.");

	do (void)fgetc(file); while (++loc<65);
	switch(type=fgetc(file)) {
	    case 0 : GetCell=GetCellS; break;
	    case 1 : GetCell=GetCellL; break;
	    default: Error("unknown data type.");
	}
	++loc;

	do (void)fgetc(file); while (++loc<527);
	if ((*fRead)(&width,sizeof(width),1,file)==1 && width>0)
	    loc+=sizeof(width);
	else
	    Error("bad image width.");

	printf("%d\n",(int)width);

	do (void)fgetc(file); while (++loc<537);
	(void)(*fRead)(&DSU,sizeof(DSU),1,file); loc+=sizeof(DSU);

	dr=DSU/(double)width;
	printf("%lf\n",dr);

	do (void)fgetc(file); while (++loc<557);
	if ((*fRead)(&I,sizeof(I),1,file)==1 && I>0)
	    loc+=sizeof(I);
	else
	    Error("bad number of I.");

	do (void)fgetc(file); while (++loc<569);
	if ((*fRead)(&I0,sizeof(I0),1,file)==1 && I0>0)
	    loc+=sizeof(I0);
	else
	    Error("bad number of I0.");
/*
	do (void)fgetc(file); while (++loc<571);
*/
	if ((*fRead)(&dark,sizeof(dark),1,file)==1 && dark>0)
	    loc+=sizeof(dark);
	else
	    Error("bad number of dark.");

	hight=(int)I+(int)I0+(int)dark;
	printf("%d\t%d\t%d\t%d\n",(int)I,(int)I0,(int)dark,hight);

	total=(int)width*hight;
	switch(type) {
	    case 0 : sData=(unsigned short *)MA(total,sData); break;
	    case 1 : lData=(unsigned long  *)MA(total,lData); break;
	}
	if (sData==NULL && lData==NULL)
	    Error("image data memory allocation error.");

	do (void)fgetc(file); while (++loc<=1024);
	switch(type) {
	    case 0 : read=(*fRead)(sData,sizeof(*sData),total,file); break;
	    case 1 : read=(*fRead)(lData,sizeof(*lData),total,file); break;
	}
	if (read!=total)
	    if (read>0)
		fputs("unexpected end of file, but continued.\n",stderr);
	    else
		Error("no image data.");

	if (file!=stdin) fclose(file);

	switch(type) {
	    case 0 :
		s=sData;
		for (y=0; y<hight; y++)
		for (x=0; x<width; x++) {
		    if (--read<0) *s=0;

		    if (s==sData)
			min=max=(*s);
		    else {
			if (*s<min) min=(*s);
			if (*s>max) max=(*s);
		    }
		    ++s;
		}
		break;
	    case 1 :
		l=lData;
		for (y=0; y<hight; y++)
		for (x=0; x<width; x++) {
		    if (--read<0) *l=0;

		    if (l==lData)
			min=max=(*l);
		    else {
			if (*l<min) min=(*l);
			if (*l>max) max=(*l);
		    }
		    ++l;
		}
		break;
	}
	printf("%lu\t%lu\n",min,max);

	lim=(1<<BPS)-1;
	if (argc==4) {
	    base=(double)min;
	    if (min==max) {
		fputs("bad image data range, but continued.\n",stderr);

		step=1.0;
	    }
	    else
		step=(double)(max-min)/(double)lim;
	}
	printf("%lf\t%lf\n",base,step);

	idx=0;
	sprintf(desc,"%lf\t%lf\t%lf",dr,base,step);

#ifdef	FUNCTION_CELL
	StoreImageFile(argv[argc-1],(int)width,hight,BPS,GetCell,desc);
#else
{
	Cell	**cell=(Cell **)MA(hight,cell);

	if (cell==NULL) Error("image cell memory allocation error.");

	for (y=0; y<hight; y++) {
	    switch(type) {
		case 0 : cell[y]=(Cell *)(sData+idx); break;
		case 1 : cell[y]=(Cell *)(lData+idx); break;
	    }
	    for (x=0; x<width; x++) cell[y][x]=(*GetCell)();
	}
	StoreImageFile(argv[argc-1],(int)width,hight,BPS,cell,desc);

	free(cell);
}
#endif
	switch(type) {
	    case 0 : free(sData); break;
	    case 1 : free(lData); break;
	}
	return 0;
}
