
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "cell.h"
#include "sif.h"

static void	Error(msg)
char		*msg;
{
	fputs(msg,stderr); fputc('\n',stderr); exit(1);
}

static size_t	fReadRev(ptr,size,items,file)
char		*ptr;
size_t		size,items;
FILE		*file;
{
	size_t	i,j,size1=size-1,size2=size/2,read=fread(ptr,size,items,file);
	char	*p1,*p2,c;

	for (i=0; i<read; i++) {
	    for (j=0; j<size2; j++) {
		p1=ptr+j; p2=ptr+size1-j; c=(*p1); *p1=(*p2); *p2=c;
	    }
	    ptr+=size;
	}
	return read;
}

static float	*data;
static int	idx,lim;
static float	base,step;

static Cell	GetCell()
{
	float	f=(data[idx++]-base)/step;

	return (f<0.0)?0:(f>=(float)lim)?(Cell)lim:(Cell)floor(f+0.5);
}

#define DESC_LEN	256
#define	MA(num,var)	malloc((size_t)(num)*sizeof(*(var)))

int	main(argc,argv)
int	argc;
char	**argv;
{
	unsigned short	tmp=0x1234;
	unsigned char	*ptr=(unsigned char *)&tmp;
	size_t		(*fRead)()=(*ptr==(tmp&0xff))?fread:fReadRev;
	FILE		*file;
	float		xCenter,yCenter,radius,CTvalue,
			dz,dFD,dOF,DSU,DSV,mf,dr,dx,dy,*f,min,max;
	int		BPS,total,read,x,y,loc=1;
	short		gap,cum,height,width;
	char		desc[DESC_LEN];

	if (argc!=4 && argc!=6 && argc!=8 && argc!=10)
	Error(
"usage : ctr2tiff CTR {xCenter yCenter radius CTvalue} {base step} BPS TIFF"
	);

	if (argv[1][0]=='-')
	    file=stdin;
	else
	    if ((file=fopen(argv[1],"rb"))==NULL) {
		fputs(argv[1],stderr); Error(" : file not open.");
	    }

	if (argc==8 || argc==10) {
	    xCenter=atof(argv[2]);
	    yCenter=atof(argv[3]);
	    if ((radius=atof(argv[4]))<=0.0) Error("bad radius of FOV.");

	    CTvalue=atof(argv[5]);
	}
	if (argc==6 || argc==10) {
	    base=atof(argv[argc-4]);
	    if ((step=atof(argv[argc-3]))<=0.0) Error("bad image data step.");
	}
	if ((BPS=atoi(argv[argc-2]))<=0 || BPS>8*sizeof(Cell))
	    Error("bad image cell BPS.");

	do (void)fgetc(file); while (++loc<409);
	(void)(*fRead)(&gap,sizeof(gap),1,file); loc+=sizeof(gap);
	(void)(*fRead)(&cum,sizeof(cum),1,file); loc+=sizeof(cum);

	do (void)fgetc(file); while (++loc<447);
	(void)(*fRead)(&dz,sizeof(dz),1,file); loc+=sizeof(dz);

	do (void)fgetc(file); while (++loc<461);
	(void)(*fRead)(&height,sizeof(height),1,file); loc+=sizeof(height);

	do (void)fgetc(file); while (++loc<527);
	if ((*fRead)(&width,sizeof(width),1,file)==1 && width>0)
	    loc+=sizeof(width);
	else
	    Error("bad image width.");

	printf("%d\n",(int)width);

/*
	do (void)fgetc(file); while (++loc<529);
*/
	(void)(*fRead)(&dFD,sizeof(dFD),1,file); loc+=sizeof(dFD);
	(void)(*fRead)(&dOF,sizeof(dOF),1,file); loc+=sizeof(dOF);
	(void)(*fRead)(&DSU,sizeof(DSU),1,file); loc+=sizeof(DSU);
	(void)(*fRead)(&DSV,sizeof(DSV),1,file); loc+=sizeof(DSV);

	if (dOF<=0.0 || (mf=dFD/dOF)<=0.0)
	    dr=0.0;
	else {
	    dr=DSU/((float)width*mf);
	    if (height>0) dz=(float)(gap+cum)*DSV/((float)height*mf);
	}
	printf("%f\n%f\n",dr,dz);

	total=(int)width*(int)width;
	if ((data=(float *)MA(total,data))==NULL)
	    Error("image data memory allocation error.");

	do (void)fgetc(file); while (++loc<=1024);
	if ((read=(*fRead)(data,sizeof(*data),total,file))!=total)
	    if (read>0)
		fputs("unexpected end of file, but continued.\n",stderr);
	    else
		Error("no image data.");

	if (file!=stdin) fclose(file);

	f=data;
	for (y=0; y<width; y++)
	for (x=0; x<width; x++) {
	    if (--read<0) *f=0.0;

	    if (argc==8 || argc==10) {
		dx=((float)x-xCenter)/radius;
		dy=((float)y-yCenter)/radius; if (dx*dx+dy*dy>1.0) *f=CTvalue;
	    }
	    if (f==data)
		min=max=(*f);
	    else {
		if (*f<min) min=(*f);
		if (*f>max) max=(*f);
	    }
	    ++f;
	}
	printf("%f\t%f\n",min,max);

	lim=(1<<BPS)-1;
	if (argc==4 || argc==8) {
	    base=min;
	    if (min==max) {
		fputs("bad image data range, but continued.\n",stderr);

		step=1.0;
	    }
	    else
		step=(max-min)/(float)lim;
	}
	printf("%f\t%f\n",base,step);

	idx=0;
	sprintf(desc,"%f\t%f\t%f",dr,base,step);

#ifdef	FUNCTION_CELL
	StoreImageFile(argv[argc-1],(int)width,(int)width,BPS,GetCell,desc);
#else
{
	Cell	**cell=(Cell **)MA(width,cell);

	if (cell==NULL) Error("image cell memory allocation error.");

	for (y=0; y<width; y++) {
	    cell[y]=(Cell *)(data+idx);

	    for (x=0; x<width; x++) cell[y][x]=GetCell();
	}
	StoreImageFile(argv[argc-1],(int)width,(int)width,BPS,cell,desc);

	free(cell);
}
#endif
	free(data); return 0;
}
