
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <dirent.h>

#ifndef TAB
#define TAB	'\t'
#endif

#ifndef	EXT
#define EXT	".CTD"
#endif

typedef	unsigned char	uChar;
typedef	unsigned short	uShort;
typedef	unsigned long	uLong;

typedef struct List {
		int		loc,type,num,size;
		char		*com,*format;
		uChar		*data;
		struct List	*next;
	} List;

#define TYPE_CHAR	0
#define TYPE_UCHAR	1
#define TYPE_SHORT	2
#define TYPE_USHORT	3
#define TYPE_LONG	4
#define TYPE_ULONG	5
#define TYPE_FLOAT	6
#define TYPE_DOUBLE	7
#define TYPE_STRING	8
#define TYPE_O		9
#define TYPE_X		10

static struct {
		char	*key;		/* keyword */
		int	size;		/* data size */
		char	*format;	/* print format */
	} rule[]={
		{"c", 1,"%d"  },	/* 0 */
		{"uc",1,"%u"  },	/* 1 */
		{"sh",2,"%d"  },	/* 2 */
		{"us",2,"%u"  },	/* 3 */
		{"l", 4,"%d"  },	/* 4 */
		{"ul",4,"%lu" },	/* 5 */
		{"f", 4,"%f"  },	/* 6 */
		{"d", 8,"%lf" },	/* 7 */
		{"st",1,"%s"  },	/* 8 */
		{"o", 1,"%03o"},	/* 9 */
		{"x", 1,"%02x"}};	/* 10 */

#define STR_LEN	256

static char	text[STR_LEN];
static int	line;

static void	Abort(msg)
char		*msg;
{
	fprintf(stderr,"%s at line %d : %s\n",msg,line,text); exit(1);
}

#define CHECK_KEY(TYPE)	\
	if (!strncasecmp(key,rule[TYPE].key,strlen(rule[TYPE].key))) type=TYPE

static List	*CheckScript()
{
	char	com[STR_LEN],*ptr,sep[STR_LEN],*tmp,key[STR_LEN];
	int	loc,type,num;
	List	*list;

	com[0]='\0';
	if ((ptr=strchr(text,'#'))==NULL)
	    ptr=text+strlen(text)-1;
	else {
	    sprintf(sep," \r\n%c",TAB);
	    if ((tmp=strtok(ptr+1,sep))!=NULL) {
		strcpy(com,tmp);
		while ((tmp=strtok(NULL,sep))!=NULL) {
		    strcat(com," "); strcat(com,tmp);
		}
	    }
	}
	do {
	    *ptr='\0'; if (--ptr<text) return NULL;
	} while (*ptr==' ' || *ptr=='\t');

	switch(sscanf(text,"%d %s %d",&loc,key,&num)) {
	    case 1  : Abort("too few script item");
	    case 2  : num=1;
	    case 3  : break;
	    default : Abort("script items scan error");
	}
	CHECK_KEY(TYPE_CHAR  ); else
	CHECK_KEY(TYPE_UCHAR ); else
	CHECK_KEY(TYPE_SHORT ); else
	CHECK_KEY(TYPE_USHORT); else
	CHECK_KEY(TYPE_LONG  ); else
	CHECK_KEY(TYPE_ULONG ); else
	CHECK_KEY(TYPE_FLOAT ); else
	CHECK_KEY(TYPE_DOUBLE); else
	CHECK_KEY(TYPE_STRING); else
	CHECK_KEY(TYPE_O     ); else
	CHECK_KEY(TYPE_X     ); else Abort("unknown data type");

	if ((list=(List *)malloc(sizeof(List)))==NULL ||
	    (list->com=(char *)malloc(strlen(com)+1))==NULL ||
	    (list->data=(uChar *)
	     malloc(rule[type].size*num+((type==TYPE_STRING)?1:0)))==NULL)
	    Abort("list memory allocation error");

	list->loc=loc; list->type=type; list->num=num; strcpy(list->com,com);
	list->size=rule[type].size; list->format=rule[type].format;
	list->next=NULL; return list;
}

#define PRINT_DATA(TYPE)	\
	for (idx=0; idx<list->num; idx++) {\
	    if (idx>0) putchar(' ');\
	    printf(list->format,*((TYPE *)(list->data+list->size*idx)));\
	}

static void	PrintCTD(dir,name,fRead,top)
char		*dir,*name;
size_t		(*fRead)();
List		*top;
{
	char	path[STR_LEN];
	FILE	*file;
	List	*list;
	int	idx;

	sprintf(path,"%s%s",dir,name);
	if ((file=fopen(path,"rb"))==NULL) {
	    fprintf(stderr,"%s : file not open.\n",path); return;
	}
	for (list=top; list!=NULL; list=list->next)
	    if (fseek(file,list->loc-1,0)!=0 ||
		(*fRead)(list->data,list->size,list->num,file)!=list->num) {
		fprintf(stderr,"%d : data read error at %d\n",path,list->loc);
		fclose(file); return;
	    }

	fclose(file);

	printf("%s",name);
	for (list=top; list!=NULL; list=list->next) {
	    putchar(TAB);
	    switch(list->type) {
		case TYPE_CHAR   : PRINT_DATA(char  ); break;
		case TYPE_UCHAR  :
		case TYPE_O      :
		case TYPE_X      : PRINT_DATA(uChar ); break;
		case TYPE_SHORT  : PRINT_DATA(short ); break;
		case TYPE_USHORT : PRINT_DATA(uShort); break;
		case TYPE_LONG   : PRINT_DATA(long  ); break;
		case TYPE_ULONG  : PRINT_DATA(uLong ); break;
		case TYPE_FLOAT  : PRINT_DATA(float ); break;
		case TYPE_DOUBLE : PRINT_DATA(double); break;
		case TYPE_STRING :
		    for (idx=0; idx<list->num; idx++)
			if (list->data[idx]==TAB) list->data[idx]=' ';

		    list->data[list->num]='\0';
		    printf(list->format,list->data);
		    break;
	    }
	}
	putchar('\n');
}

static size_t	fReadRev(ptr,size,items,file)
char		*ptr;
size_t		size,items;
FILE		*file;
{
	size_t	i,j,size1=size-1,size2=size/2,read=fread(ptr,size,items,file);
	char	*p1,*p2,c;

	for (i=0; i<read; i++) {
	    for (j=0; j<size2; j++) {
		p1=ptr+j; p2=ptr+size1-j; c=(*p1); *p1=(*p2); *p2=c;
	    }
	    ptr+=size;
	}
	return read;
}

int	main(argc,argv)
int	argc;
char	**argv;
{
	uShort		tmp=0x1234;
	uChar		*ptr=(uChar *)&tmp;
	size_t		(*fRead)()=(*ptr==(tmp&0xff))?fread:fReadRev;
	FILE		*script;
	DIR		*dir;
	List		*list,*top=NULL,**next=(&top);
	struct dirent	*ent;
	int		len,lim=strlen(EXT);

	if (argc!=2 && argc!=3) {
	    fprintf(stderr,"usage : %s {script} directory\n",*argv); exit(1);
	}

	if (argc==2)
	    script=stdin;
	else
	    if ((script=fopen(argv[1],"r"))==NULL) {
		fprintf(stderr,"%s : file not open.\n",argv[1]); exit(1);
	    }

	if ((dir=opendir(argv[argc-1]))==NULL) {
	    fprintf(stderr,"%s : directory not open.\n",argv[argc-1]); exit(1);
	}

	for (line=1; fgets(text,STR_LEN,script)!=NULL; line++)
	    if (text[strlen(text)-1]!='\n') {
		fprintf(stderr,"too long script at line %d\n",line); exit(1);
	    }
	    else
		if ((*next=CheckScript())!=NULL) next=(&((*next)->next));

	if (script!=stdin) fclose(script);

	for (list=top; list!=NULL; list=list->next) {
	    putchar(TAB); printf("%s",list->com);
	}
	putchar('\n');

	while ((ent=readdir(dir))!=NULL)
	    if ((len=strlen(ent->d_name))>lim &&
		!strncasecmp(EXT,ent->d_name+len-lim,lim))
		PrintCTD(argv[argc-1],ent->d_name,fRead,top);

	closedir(dir); return 0;
}
