
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <dirent.h>
#include <math.h>
#include "cell.h"
#include "sif.h"

static void	Error(msg)
char		*msg;
{
	fputs(msg,stderr); fputc('\n',stderr); exit(1);
}

static size_t	fReadRev(ptr,size,items,file)
char		*ptr;
size_t		size,items;
FILE		*file;
{
	size_t	i,j,size1=size-1,size2=size/2,read=fread(ptr,size,items,file);
	char	*p1,*p2,c;

	for (i=0; i<read; i++) {
	    for (j=0; j<size2; j++) {
		p1=ptr+j; p2=ptr+size1-j; c=(*p1); *p1=(*p2); *p2=c;
	    }
	    ptr+=size;
	}
	return read;
}

static float	*data;
static int	idx,lim;
static float	base,step;

static Cell	GetCell()
{
	float	f=(data[idx++]-base)/step;

	return (f<0.0)?0:(f>=(float)lim)?(Cell)lim:(Cell)floor(f+0.5);
}

#define PATH_LEN	256
#define DESC_LEN	256
#define	MA(num,var)	malloc((size_t)(num)*sizeof(*(var)))

int	main(argc,argv)
int	argc;
char	**argv;
{
	unsigned short	tmp=0x1234;
	unsigned char	*ptr=(unsigned char *)&tmp;
	size_t		(*fRead)()=(*ptr==(tmp&0xff))?fread:fReadRev;
	DIR		*dir;
	float		xCenter,yCenter,radius,CTvalue,
			dz,dFD,dOF,DSU,DSV,mf,dr,dx,dy,*f,min,max;
	int		BPS,len,total,read,x,y;
	struct dirent	*ent;
	char		path[PATH_LEN],desc[DESC_LEN],name[PATH_LEN];
	FILE		*file;
	short		gap,cum,height,width;

	if (argc!=4 && argc!=6 && argc!=8 && argc!=10)
	Error(
"usage : mctr2tif CTR {xCenter yCenter radius CTvalue} {base step} BPS TIFF"
	);

	if ((dir=opendir(argv[1]))==NULL) {
	    fputs(argv[1],stderr); Error(" : directory not open.");
	}
	if (argc==8 || argc==10) {
	    xCenter=atof(argv[2]);
	    yCenter=atof(argv[3]);
	    if ((radius=atof(argv[4]))<=0.0) Error("bad radius of FOV.");

	    CTvalue=atof(argv[5]);
	}
	if (argc==6 || argc==10) {
	    base=atof(argv[argc-4]);
	    if ((step=atof(argv[argc-3]))<=0.0) Error("bad image data step.");
	}
	if ((BPS=atoi(argv[argc-2]))<=0 || BPS>8*sizeof(Cell))
	    Error("bad image cell BPS.");

	while ((ent=readdir(dir))!=NULL)
	    if ((len=strlen(ent->d_name))>4 &&
		ent->d_name[len-4]=='.' &&
	       (ent->d_name[len-3]|0x20)=='c' &&
	       (ent->d_name[len-2]|0x20)=='t' &&
	       (ent->d_name[len-1]|0x20)=='r')
{
	sprintf(path,"%s%s",argv[1],ent->d_name);
	if ((file=fopen(path,"rb"))==NULL) {
	    fprintf(stderr,"%s : file not open.\n",path); continue;
	}
	fseek(file,408L,SEEK_SET);
	(void)(*fRead)(&gap,sizeof(gap),1,file);
	(void)(*fRead)(&cum,sizeof(cum),1,file);

	fseek(file,446L,SEEK_SET);
	(void)(*fRead)(&dz,sizeof(dz),1,file);

	fseek(file,460L,SEEK_SET);
	(void)(*fRead)(&height,sizeof(height),1,file);

	fseek(file,526L,SEEK_SET);
	if ((*fRead)(&width,sizeof(width),1,file)!=1 || width<=0) {
	    fclose(file);
	    fprintf(stderr,"%s : bad image width.\n",path); continue;
	}

/*
	fseek(file,528L,SEEK_SET);
*/
	(void)(*fRead)(&dFD,sizeof(dFD),1,file);
	(void)(*fRead)(&dOF,sizeof(dOF),1,file);
	(void)(*fRead)(&DSU,sizeof(DSU),1,file);
	(void)(*fRead)(&DSV,sizeof(DSV),1,file);

	if (dOF<=0.0 || (mf=dFD/dOF)<=0.0)
	    dr=0.0;
	else {
	    dr=DSU/((float)width*mf);
	    if (height>0) dz=(float)(gap+cum)*DSV/((float)height*mf);
	}

	total=(int)width*(int)width;
	if ((data=(float *)MA(total,data))==NULL)
	    Error("image data memory allocation error.");

	fseek(file,1024L,SEEK_SET);
	if ((read=(*fRead)(data,sizeof(*data),total,file))!=total)
	    if (read>0)
		fprintf(stderr,"%s : unexpected end of file, but continued.\n",
			path);
	    else {
		fclose(file); free(data);
		fprintf(stderr,"%s : no image data.\n",path); continue;
	    }

	fclose(file);

	f=data;
	for (y=0; y<width; y++)
	for (x=0; x<width; x++) {
	    if (--read<0) *f=0.0;

	    if (argc==8 || argc==10) {
		dx=((float)x-xCenter)/radius;
		dy=((float)y-yCenter)/radius; if (dx*dx+dy*dy>1.0) *f=CTvalue;
	    }
	    if (f==data)
		min=max=(*f);
	    else {
		if (*f<min) min=(*f);
		if (*f>max) max=(*f);
	    }
	    ++f;
	}

	lim=(1<<BPS)-1;
	if (argc==4 || argc==8) {
	    base=min;
	    if (min==max) {
		fprintf(stderr,"%s : bad image data range, but continued.\n",
			path);

		step=1.0;
	    }
	    else
		step=(max-min)/(float)lim;
	}

	strncpy(name,ent->d_name,len-4);
	printf("%s\t%d\t%f\t%f\t%f\t%f\t%f\t%f\n",
	       name,(int)width,dr,dz,min,max,base,step);

	idx=0;
	sprintf(desc,"%f\t%f\t%f",dr,base,step);
	sprintf(path,"%s%s.tif",argv[argc-1],name);

#ifdef	FUNCTION_CELL
	StoreImageFile(path,(int)width,(int)width,BPS,GetCell,desc);
#else
{
	Cell	**cell=(Cell **)MA(width,cell);

	if (cell==NULL) Error("image cell memory allocation error.");

	for (y=0; y<width; y++) {
	    cell[y]=(Cell *)(data+idx);

	    for (x=0; x<width; x++) cell[y][x]=GetCell();
	}
	StoreImageFile(path,(int)width,(int)width,BPS,cell,desc);

	free(cell);
}
#endif
	free(data);
}
	closedir(dir); return 0;
}
