
#include <stdio.h>
#include <stdlib.h>
#include "CBP.h"
#include "sri.h"

static void	Error(msg)
char		*msg;
{
	fputs(msg,stderr); fputc('\n',stderr); exit(1);
}

#define LINE_LEN	512
#define DESC_LEN	256
#define RADIAN		0.01745329251994329576

#define R0_STR_LEN	128
#define R0_STR_FORMAT	"%19.18lf"

int	main(argc,argv)
int	argc;
char	**argv;
{
	double	dr,r0,t0,f0,df,p,f1,f2;
	int	bps,N,M,n,m,v,h;
	FILE	*file;
	char	line[LINE_LEN],desc[DESC_LEN],str[R0_STR_LEN],*ptr;
	CBP	*cbp;
	Float	**f;

	if (argc!=7 && argc!=9)
	    Error("usage : p2tiff XP dr r0/dr t0 {f0 df} BPS TIFF");

	if ((dr=atof(argv[2]))<=0.0) Error("bad detector interval.");

	r0=atof(argv[3])*dr;
	t0=atof(argv[4]);

	if (argc==9) {
	    f0=atof(argv[5]);
	    if ((df=atof(argv[6]))<=0.0) Error("bad CT-value step.");
	}
	if ((bps=atoi(argv[argc-2]))<=0) Error("bad CT-image BPS.");

	if (argv[1][0]=='-')
	    file=stdin;
	else
	    if ((file=fopen(argv[1],"rb"))==NULL)
		Error("X-ray projection data file not found.");

	if (fgets(line,LINE_LEN,file)==NULL ||
	    sscanf(line,"%d %d",&N,&M)!=2 ||
	    N<=0 || M<=0) Error("bad number of X-ray projections.");

	if ((cbp=SetUpCBP(N,M,dr,r0,t0*RADIAN))==NULL)
	    Error("SetUpCBP() memory allocation error.");

	for (m=0; m<M; m++)
	for (n=0; n<N; n++) {
	    fread(&p,sizeof(p),1,file); cbp->pq[m][n]=p;
	}
	if (file!=stdin) fclose(file);

	Convolution(cbp);

	if (!ReGridingCP(cbp)) Error("ReGridingCP() memory allocation error.");

	if ((f=BPImage(cbp))==NULL) Error("BPImage() memory allocation error.");

	f1=f2=f[0][0];
	for (v=0; v<N; v++)
	for (h=0; h<N; h++) if (f[v][h]<f1) f1=f[v][h]; else
			    if (f[v][h]>f2) f2=f[v][h];

	printf("%lf\t%lf\n",f1,f2);

	sprintf(str,R0_STR_FORMAT,r0); for (ptr=str; *ptr!='\0'; ptr++) ;
				       while (*(--ptr)=='0') *ptr='\0';
				       if (*ptr=='.') ptr[1]='0';

	if (argc==9)
	    sprintf(desc,"%s\t%s\n%d\t%s\n%lf\t%lf\n%s\t%s",
			 argv[2],str,M,argv[4],f1,f2,argv[5],argv[6]);
	else {
	    f0=f1;
	    df=(f2-f1)/(double)((1<<bps)-1);

	    sprintf(desc,"%s\t%s\n%d\t%s\n%lf\t%lf",
			 argv[2],str,M,argv[4],f1,f2);
	}
	StoreReconstructImage(argv[argc-1],N,f,f0-df/2.0,df,bps,desc);

	return 0;
}
